from __future__ import annotations

from abc import abstractmethod
from math import pi
from typing import TYPE_CHECKING, Any, Callable

from numpy.typing import NDArray
from tqdm import tqdm

from cratermaker.constants import FloatLike
from cratermaker.core.crater import Crater
from cratermaker.utils.component_utils import ComponentBase, import_components

if TYPE_CHECKING:
    from cratermaker.components.surface import Surface, SurfaceView


class Morphology(ComponentBase):
    def __init__(self, **kwargs: Any) -> None:
        """
        Initialize the Morphology class.

        Parameters
        ----------
        **kwargs : Any
            Additional keyword arguments.

        Raises
        -------

        TypeError
            If the crater is not an instance of Crater.

        """
        super().__init__(**kwargs)
        self._queue_manager: CraterQueueManager | None = None

    def __str__(self) -> str:
        base = super().__str__()
        return base

    @classmethod
    def maker(
        cls,
        morphology: str | type[Morphology] | Morphology | None = None,
        **kwargs: Any,
    ) -> Morphology:
        """
        Initialize a component model with the given name or instance.

        Parameters
        ----------
        morphology : str or Morphology or None
            The name of the morphology model to use, or an instance of Morphology. If None, the default "simplemoon" is used.
        **kwargs : Any
            Additional keyword arguments that are required for the specific morphology model being created.

        Returns
        -------
        component
            An instance of the specified component model.

        Raises
        ------
        KeyError
            If the specified morphology model name is not found in the registry.
        TypeError
            If the specified morphology model is not a string or a subclass of Morphology.
        """

        # Call the base class version of make and pass the morphology argument as the component argument
        if morphology is None:
            morphology = "simplemoon"
        morphology = super().maker(component=morphology, **kwargs)
        return morphology

    def emplace(self, crater: Crater, surface: Surface, **kwargs: Any) -> None:
        """
        Convenience method to immediately emplace a crater onto the surface.
        Initializes and uses the queue system behind the scenes.

        Parameters
        ----------
        crater : Crater
            The crater to be emplaced.
        surface : Surface
            The surface object to modify.
        **kwargs : Any
            Additional keyword arguments.
        """
        if self._queue_manager is None:
            self.init_queue_manager(surface)
        self.enqueue_crater(crater)
        self.process_queue(surface)

    def form_crater(self, crater: Crater, surface: Surface, **kwargs: Any) -> Surface:
        """
        This method forms the interior of the crater by altering the elevation variable of the surface mesh.

        Parameters
        ----------
        crater : Crater
            The crater object to be formed.
        surface : Surface
            The surface to be altered. This will be modified in place.
        **kwargs : dict
            Additional keyword arguments to be passed to internal functions.

        Returns
        -------
        None
        """
        from cratermaker.components.surface import Surface

        if not isinstance(surface, Surface):
            raise TypeError("surface must be an instance of Surface")
        if not isinstance(crater, Crater):
            raise TypeError("crater must be an instance of Crater")

        # Find the node and face center of the crater
        self.node_index, self.face_index = surface.find_nearest_index(crater.location)

        # Test if the ejecta is big enough to modify the surface
        ejecta_rmax = self.rmax(
            crater, minimum_thickness=surface.smallest_length, feature="ejecta"
        )
        ejecta_region_view = surface.extract_region(crater.location, ejecta_rmax)
        ejecta_area = pi * ejecta_rmax**2
        if (
            ejecta_region_view is None
            or ejecta_area < surface.face_areas[self.face_index]
        ):  # The crater is too small to change the surface
            return

        crater_rmax = self.rmax(
            crater, minimum_thickness=surface.smallest_length, feature="crater"
        )
        crater_region_view = surface.extract_region(crater.location, crater_rmax)
        if (
            crater_region_view is not None
        ):  # The crater is big enough to affect the surface
            crater_area = pi * crater_rmax**2

            # Check to make sure that the face at the crater location is not smaller than the crater area
            if crater_area > surface.face_areas[self.face_index]:
                # Form the crater shape
                surface = self.crater_shape(crater, crater_region_view, surface)

        # Now form the ejecta blanket
        surface = self.ejecta_shape(crater, ejecta_region_view, surface)
        return

    def affected_indices(
        self, crater: Crater, surface: Surface
    ) -> tuple[set[int], set[int]]:
        """
        Determine the set of node and face indices affected by a crater's ejecta blanket.

        Parameters
        ----------
        crater : Crater
            The crater object whose effect region is to be computed.
        surface : Surface
            The surface object on which the crater would be emplaced.

        Returns
        -------
        affected_nodes : set of int
            The set of node indices affected by the crater.
        affected_faces : set of int
            The set of face indices affected by the crater.
        """
        rmax = self.rmax(
            crater, minimum_thickness=surface.smallest_length, feature="ejecta"
        )
        region_view = surface.extract_region(crater.location, rmax)
        if region_view is None:
            return set(), set()
        return set(region_view.node_indices), set(region_view.face_indices)

    def init_queue_manager(self, surface: Surface) -> None:
        """
        Initialize the crater queue manager with a surface-dependent overlap function.

        Parameters
        ----------
        surface : Surface
            The surface object used to determine which nodes and faces each crater affects.
        """

        def overlap_fn(crater: Crater) -> tuple[set[int], set[int]]:
            return self.affected_indices(crater, surface)

        self._queue_manager = CraterQueueManager(overlap_fn)

    def enqueue_crater(
        self,
        crater: Crater | None = None,
        surface: "Surface" | None = None,
        **kwarg: Any,
    ) -> None:
        """
        Add a crater to the queue for later emplacement. Automatically initializes
        the queue manager if it hasn't been set.

        Parameters
        ----------
        crater : Crater, optional
            The crater object to enqueue. If None, one is created from keyword args.
        surface : Surface, optional
            The surface used to determine overlap regions if initialization is needed.
        **kwarg : Any
            Additional keyword arguments for crater construction.

        Raises
        ------
        RuntimeError
            If the queue manager must be initialized but no surface is provided.
        """
        if self._queue_manager is None:
            if surface is None:
                raise RuntimeError(
                    "Surface must be provided to initialize queue manager."
                )
            self.init_queue_manager(surface)

        if crater is None:
            crater = Crater.maker(**kwarg)
        self._queue_manager.push(crater)

    def process_queue(self, surface: Surface) -> None:
        """
        Process all queued craters in the order they were added, forming non-overlapping
        batches and applying each to the surface.

        Parameters
        ----------
        surface : Surface
            The surface object to be modified in place.

        Raises
        ------
        RuntimeError
            If the queue manager has not been initialized.
        """
        if not hasattr(self, "_queue_manager"):
            raise RuntimeError(
                "Queue manager has not been initialized. Call init_queue_manager first."
            )

        def _batch_process(pbar=None):
            while not self._queue_manager.is_empty():
                batch = self._queue_manager.peek_next_batch()
                for crater in batch:
                    self.form_crater(crater, surface)
                    if pbar is not None:
                        pbar.update(1)
                self._queue_manager.pop_batch(batch)
                self._queue_manager.clear_active()
            return

        total_craters = len(self._queue_manager._queue)
        if total_craters > 10:
            with tqdm(total=total_craters, desc="Processing craters") as pbar:
                _batch_process(pbar)
        else:
            _batch_process()
        return

    @abstractmethod
    def crater_shape(
        self,
        crater: Crater,
        region_view: SurfaceView | NDArray,
        surface: Surface | NDArray,
    ) -> Surface | NDArray: ...

    @abstractmethod
    def ejecta_shape(
        self,
        crater: Crater,
        region_view: SurfaceView | NDArray,
        surface: Surface | NDArray,
    ) -> Surface | NDArray: ...

    @abstractmethod
    def rmax(
        self,
        crater: Crater,
        minimum_thickness: FloatLike,
        feature: str = "ejecta",
    ) -> float: ...


class CraterQueueManager:
    """
    A manager for craters awaiting emplacement. Craters are processed in order (FIFO),
    but batches of non-overlapping craters can be processed simultaneously.

    Parameters
    ----------
    overlap_fn : Callable[[Crater], tuple[set[int], set[int]]]
        A function that takes a crater and returns a tuple of (node indices, face indices) affected.
    """

    def __init__(self, overlap_fn: Callable[[Crater], tuple[set[int], set[int]]]):
        self._queue: list[Crater] = []
        self._active_entities: set[int] = set()
        self._overlap_fn = overlap_fn

    def push(self, crater: Crater) -> None:
        self._queue.append(crater)

    def peek_next_batch(self) -> list["Crater"]:
        """
        Return a list of the next batch of craters that do not overlap with each other
        or the current active region.
        """
        batch = []
        reserved = set(self._active_entities)
        for crater in self._queue:
            node_indices, face_indices = self._overlap_fn(crater)
            combined = node_indices | face_indices
            if reserved.isdisjoint(combined):
                batch.append(crater)
                reserved.update(combined)
            else:
                break
        return batch

    def pop_batch(self, batch: list["Crater"]) -> None:
        """
        Remove a processed batch of craters from the queue.
        """
        for crater in batch:
            self._queue.remove(crater)
            node_indices, face_indices = self._overlap_fn(crater)
            self._active_entities.update(node_indices | face_indices)

    def clear_active(self) -> None:
        """
        Clear the active region set after batch processing is complete.
        """
        self._active_entities.clear()

    def is_empty(self) -> bool:
        return len(self._queue) == 0


import_components(__name__, __path__, ignore_private=True)



from typing import Any
from warnings import warn

from numpy.random import Generator

from cratermaker.components.projectile import Projectile
from cratermaker.components.target import Target
from cratermaker.constants import FloatLike


@Projectile.register("asteroids")
class AsteroidProjectiles(Projectile):
    _catalogue = {
        "Mercury": 41100.0,
        "Venus": 29100.0,
        "Earth": 24600.0,
        "Moon": 22100.0,
        "Mars": 10700.0,
        "Ceres": 5300.0,
        "Vesta": 5300.0,
        "MBA": 5300.0,
    }

    def __init__(
        self,
        sample: bool = True,
        density: FloatLike | None = None,
        angle: FloatLike | None = None,
        direction: FloatLike | None = None,
        location: tuple[FloatLike, FloatLike] | None = None,
        target: Target | str | None = None,
        rng: Generator | None = None,
        rng_seed: int | None = None,
        rng_state: dict | None = None,
        **kwargs: Any,
    ):
        """
        An operations class for computing the projectile properties of an asteroid source population.

        Parameters
        ----------
        sample : bool
            Flag that determines whether to sample impact velocities, angles, and directions from distributions. If set to False, the projectile velocity will be the mean velocity for the given target body.
        density : float, optional
            The density of the projectile in kg/m^3. Default is 2250 kg/m^3.
        angle : float, optional
            The impact angle in degrees. Default is 90.0 degrees (vertical impact) if `sample` is False. If `sample` is True, this value is ignored.
        direction : float | None
            The impact direction in degrees. Default is 0.0 degrees (due North) if `sample` is False. If `sample` is True, this value is ignored.`
        location : tuple[float, float] | None
            The impact location as a tuple of (longitude, latitude) in degrees. Default is (0.0, 0.0) (the equator and prime meridian) if `sample` is False. If `sample` is True, this value is ignored.
        target : Target or str.
            The name of the target body for the impact. Default is "Moon"
        rng : numpy.random.Generator | None
            A numpy random number generator. If None, a new generator is created using the rng_seed if it is provided.
        rng_seed : Any type allowed by the rng_seed argument of numpy.random.Generator, optional
            The rng_rng_seed for the RNG. If None, a new RNG is created.
        rng_state : dict, optional
            The state of the random number generator. If None, a new state is created.
        **kwargs : Any
            Additional keyword arguments.
        """
        if density is None:
            density = 2250.0
        kwargs.pop("mean_velocity", None)
        kwargs.pop("velocity", None)
        self._target = Target.maker(target, **kwargs)
        mean_velocity = self._set_mean_velocity()
        super().__init__(
            sample=sample,
            mean_velocity=mean_velocity,
            velocity=mean_velocity,
            angle=angle,
            direction=direction,
            location=location,
            density=density,
            target=target,
            rng=rng,
            rng_seed=rng_seed,
            rng_state=rng_state,
            **kwargs,
        )

    def __str__(self) -> str:
        base = super().__str__()
        return f"{base}\nTarget: {self.target.name}\n"

    def _set_mean_velocity(self):
        """
        Sets the mean velocity of the projectile in m/s based on the target body.

        Returns
        -------
        float
        """

        if self.target.name in self.__class__._catalogue:
            pmv = float(self.__class__._catalogue[self.target.name])
        else:
            warn(
                f"Target {self.target.name} not found in known targets. Known targets include {list(self.__class__._catalogue.keys())}. Defaulting to the Moon."
            )
            pmv = float(self.__class__._catalogue["Moon"])
        return pmv



from pathlib import Path
from typing import Any

import numpy as np
from numpy.typing import NDArray
from scipy.interpolate import interp1d
from scipy.spatial.transform import Rotation as R

from cratermaker.components.morphology import Morphology
from cratermaker.components.scaling import Scaling
from cratermaker.components.surface import Surface
from cratermaker.components.target import Target
from cratermaker.constants import FloatLike, PairOfFloats
from cratermaker.core.crater import Crater
from cratermaker.utils.general_utils import (
    format_large_units,
    parameter,
    validate_and_normalize_location,
)


@Surface.register("hireslocal")
class HiResLocalSurface(Surface):
    """
    Create a uniform grid configuration with the given pixel size.

    Parameters
    ----------
    pix : FloatLike
        The approximate face size inside the local region in meters.
    local_radius : FloatLike
        The radius of the local region in meters.
    local_location : PairOfFloats
        The longitude and latitude of the location in degrees.
    superdomain_scale_factor : FloatLike, optional
        A factor defining the ratio of cell size to the distance from the local boundary. This is set so that smallest craters
        that are modeled outside the local region are those whose ejecta could just reach the boundary. If not provide, then it will
        be computed based on a provided (or default) scaling and morphology model
    target : Target, optional
        The target body or name of a known target body for the impact simulation. If none provide, it will be either set to the default, or extracted from the scaling model if it is provied
    scaling : Scaling, optional
        A scaling model or the name of a scaling model for the impact simulation. This is only needed to set the superdomain_scale_factor if it is not specified
    morphology : Morphology, optional
        A morphology model or the name of a morphology model for the impact simulation. This is only needed to set the superdomain_scale_factor if it is not specified
    reset : bool, optional
        Flag to indicate whether to reset the surface. Default is True.
    regrid : bool, optional
        Flag to indicate whether to regrid the surface. Default is False.
    simdir : str | Path
        The main project simulation directory. Defaults to the current working directory if None.

    Returns
    -------
    HiResLocalSurface
        An instance of the HiResLocalSurface clas initialized with the given point distribution
    """

    def __init__(
        self,
        pix: FloatLike,
        local_radius: FloatLike,
        local_location: PairOfFloats,
        superdomain_scale_factor: FloatLike | None = None,
        target: Target | str | None = None,
        scaling: Scaling | str | None = None,
        morphology: Morphology | str | None = None,
        reset: bool = False,
        regrid: bool = False,
        simdir: str | Path | None = None,
        **kwargs: Any,
    ):
        super().__init__(target=target, simdir=simdir, **kwargs)
        if target is None:
            if scaling is not None and isinstance(scaling, Scaling):
                target = scaling.target
        self.pix = pix
        self.local_radius = local_radius
        self.local_location = local_location
        if superdomain_scale_factor is None:
            # Determine the scale factor for the superdomain based on the smallest crater whose ejecta can reach the edge of the
            # superdomain. This will be used to set the superdomain scale factor.
            scaling = Scaling.maker(scaling, target=self.target, **kwargs)
            morphology = Morphology.maker(morphology, target=self.target, **kwargs)
            projectile_velocity = scaling.projectile.mean_velocity * 10

            for final_diameter in np.logspace(
                np.log10(self.target.radius * 2),
                np.log10(self.target.radius / 1e6),
                1000,
            ):
                crater = Crater.maker(
                    final_diameter=final_diameter,
                    angle=90.0,
                    scaling=scaling,
                    projectile_velocity=projectile_velocity,
                    **kwargs,
                )
                rmax = morphology.rmax(crater=crater, minimum_thickness=1e-3)
                if rmax < self.target.radius * 2 * np.pi:
                    superdomain_scale_factor = rmax / crater.final_radius
                    break

        self.superdomain_scale_factor = superdomain_scale_factor
        self.load_from_files(reset=reset, regrid=regrid, **kwargs)

        return

    def __str__(self) -> str:
        base = super().__str__()
        pix = format_large_units(self.pix, quantity="length")
        local_radius = format_large_units(self.local_radius, quantity="length")
        return (
            f"{base}\n"
            f"Pixel Size: {pix}\n"
            f"Local Radius: {local_radius}\n"
            f"Local Location: ({self.local_location[0]:.2f}°, {self.local_location[1]:.2f}°)\n"
            f"Superdomain Scale Factor: {self.superdomain_scale_factor:.2f}"
        )

    @property
    def _hashvars(self):
        """
        The variables used to generate the hash.
        """
        return [
            self._component_name,
            self.radius,
            self.pix,
            self.local_radius,
            self.local_location,
            self.superdomain_scale_factor,
        ]

    def _generate_variable_size_array(self) -> tuple[NDArray, NDArray, NDArray]:
        """
        Create an array of target pixel sizes for pairs of longitude and latitude values for a high-resolution local mesh around a
        given location, with adaptive cell sizes outside the local region.


        Returns
        -------
        pix_array: ndarray
            m x n array of pixel sizes
        lon : ndarray
            longitude in degrees (length n and between -180 and 180)
        lat : ndarray
            latitude in degrees (length m and between -90 and 90)
        """

        def _pix_func(lon, lat):
            lon_rad = np.radians(lon)
            lat_rad = np.radians(lat)
            # This will be rotated into the correct position later
            loc_lon_rad = 0.0
            loc_lat_rad = 0.0

            # Calculate distance from the location to the grid point
            distance = self.calculate_haversine_distance(
                loc_lon_rad, loc_lat_rad, lon_rad, lat_rad
            )
            ans = np.where(
                distance <= self.local_radius,
                self.pix,
                (distance - self.local_radius) / self.superdomain_scale_factor
                + self.pix,
            )
            return ans

        # Suppose we know pix(lat, lon)
        # Step 1: Construct a fine preliminary grid to estimate integrals

        Lat = np.linspace(-90.0, 90.0, 733)
        Lon = np.linspace(-180.0, 180.0, 733)
        LAT, LON = np.meshgrid(Lat, Lon, indexing="ij")

        pix_values = _pix_func(LON, LAT)  # Evaluate pix on this fine grid
        w = 1.0 / pix_values

        # Step 2: Integrate w over longitude to get W_lat(lat)
        W_lat_vals = np.trapezoid(w, x=Lon, axis=1)  # integrate along lon dimension
        W_lat_cumulative = np.cumsum(W_lat_vals)
        W_lat_cumulative -= W_lat_cumulative[0]  # normalize from 0 to 1
        W_lat_cumulative /= W_lat_cumulative[-1]

        # Create a function to invert W_lat(lat)
        f_lat = interp1d(
            W_lat_cumulative, Lat, bounds_error=False, fill_value="extrapolate"
        )

        M = int(2 * np.pi * self.radius / pix_values.max()) - 1
        while M > 0:
            badval = False

            # Step 3: For each lat interval, choose lon lines similarly
            N = M + 1  # number of lon lines
            lat_lines = f_lat(np.linspace(0, 1, N))
            lon_lines = np.zeros((M, N))

            for i in range(M):
                lat_low, lat_high = lat_lines[i], lat_lines[i + 1]
                # Extract w in this lat band
                mask = (LAT >= lat_low) & (LAT <= lat_high)
                w_band = w[mask].reshape(-1, len(Lon))
                # Integrate this band over lat
                w_band_vals = np.trapezoid(
                    w_band, x=Lat[(Lat >= lat_low) & (Lat <= lat_high)], axis=0
                )
                W_lon_band_cumulative = np.cumsum(w_band_vals)
                if W_lon_band_cumulative[-1] > 0:
                    W_lon_band_cumulative -= W_lon_band_cumulative[0]
                    W_lon_band_cumulative /= W_lon_band_cumulative[-1]
                    f_lon = interp1d(
                        W_lon_band_cumulative,
                        Lon,
                        bounds_error=False,
                        fill_value="extrapolate",
                    )
                    lon_lines[i, :] = f_lon(np.linspace(0, 1, N))
                else:
                    badval = True
                    M = M // 2
                    break

            if not badval:
                break

        if badval:
            raise ValueError(
                "Could not generate a grid with the given parameters. Please try again with different parameters."
            )

        LAT = np.zeros((M, N))
        LON = np.zeros((M, N))

        for i in range(M):
            LAT[i, :] = lat_lines[
                i
            ]  # every point on this horizontal line has the same lat

        LON = lon_lines

        pix_array = _pix_func(LON, LAT)

        return pix_array, LON, LAT

    def _rotate_point_cloud(self, points):
        """
        Rotate a point cloud so that the point at [-r,0,0] moves to (lon,lat)
        using the convention:
        - Longitude [-180, 180] degrees, increasing eastward
        - Latitude [-90, 90] degrees, increasing northward
        - (0,0) lon/lat corresponds to [r,0,0] in Cartesian space

        Parameters:
            points (np.ndarray): Nx3 array of (x,y,z) points.
            lon (float): Target longitude in degrees.
            lat (float): Target latitude in degrees.
            r (float): Radius of the sphere (default=1).

        Returns:
            np.ndarray: Rotated Nx3 point cloud.
        """

        # Convert target lon, lat to radians
        lon_rad, lat_rad = (
            np.radians(self.local_location[0]),
            np.radians(self.local_location[1]),
        )

        # Compute target unit vector (correcting for lon,lat convention)
        target = np.array(
            [
                self.radius * np.cos(lat_rad) * np.cos(lon_rad),
                self.radius * np.cos(lat_rad) * np.sin(lon_rad),
                self.radius * np.sin(lat_rad),
            ]
        )

        # Original vector (the point we want to move)
        original = np.array([-self.radius, 0, 0])  # Starts at [-r, 0, 0]
        if np.isclose(lon_rad, 0.0) and np.isclose(lat_rad, 0.0):
            rotation = R.from_euler(
                "z", 180, degrees=True
            )  # 180-degree rotation around z-axis
            return rotation.apply(points)

        # Compute the axis of rotation (cross product)
        axis = np.cross(original, target)
        axis_norm = np.linalg.norm(axis)

        # If the axis is zero (no rotation needed), return the original points
        if axis_norm < 1e-10:
            return points

        axis /= axis_norm  # Normalize the axis

        # Compute the rotation angle (dot product)
        angle = np.arccos(
            np.clip(np.dot(original / self.radius, target / self.radius), -1.0, 1.0)
        )  # Normalize for dot product

        # Create rotation object
        rotvec = axis * angle  # Convert to rotation vector
        rotation = R.from_rotvec(rotvec)  # Create rotation from axis-angle

        return rotation.apply(points)

    def generate_face_distribution(self, **kwargs: Any) -> NDArray:
        """
        Creates the points that define the mesh centers.

        Returns
        -------
        (3,n) ndarray of np.float64
            Array of points on a unit sphere.

        """

        print("Generating a mesh with variable resolution faces")
        print(f"Center of local region: {self.local_location}")
        print(f"Size of local region: {self.local_radius:.2f} m")
        print(f"Hires region pixel size: {self.pix:.2f} m")
        print(
            f"Lores region pixel size: {self.pix * self.superdomain_scale_factor:.2f} m"
        )
        pix_array, lon, lat = self._generate_variable_size_array()

        points = []
        n = lon.shape[0]
        m = lat.shape[1]

        for i in range(n - 1):
            for j in range(m - 1):
                lon_range = (lon[i, j], lon[i, j + 1])
                lat_range = (lat[i, j], lat[i + 1, j])
                p = self._distribute_points(
                    distance=pix_array[i, j] / self.radius,
                    lon_range=lon_range,
                    lat_range=lat_range,
                )
                if p is not None:
                    points.append(p)

        points = np.concatenate(points, axis=1)
        points = self._rotate_point_cloud(points.T).T

        return points

    @parameter
    def pix(self):
        """
        The approximate face size for a cell of the mesh.
        """
        return self._pix

    @pix.setter
    def pix(self, value: FloatLike):
        if (
            not isinstance(value, FloatLike)
            or np.isnan(value)
            or np.isinf(value)
            or value <= 0
        ):
            raise TypeError("pix must be a positive float")
        self._pix = value

    @parameter
    def local_radius(self):
        """
        The radius of the local region in meters.
        """
        return self._local_radius

    @local_radius.setter
    def local_radius(self, value: FloatLike):
        if (
            not isinstance(value, FloatLike)
            or np.isnan(value)
            or np.isinf(value)
            or value <= 0
        ):
            raise TypeError("local_radius must be a positive float")
        if value > 2 * np.pi * self.radius:
            raise ValueError(
                "local_radius must be less than 2 * pi * radius of the target body"
            )
        self._local_radius = value

    @parameter
    def local_location(self):
        """
        The longitude and latitude of the location in degrees.
        """
        return self._local_location

    @local_location.setter
    def local_location(self, value: PairOfFloats):
        if not isinstance(value, tuple) or len(value) != 2:
            raise TypeError("local_location must be a tuple of two floats")
        self._local_location = validate_and_normalize_location(value)

    @parameter
    def superdomain_scale_factor(self):
        """
        A factor defining the ratio of cell size to the distance from the local boundary. This is set so that smallest craters that are
        modeled outside the local region are those whose ejecta could just reach the boundary.
        """
        return self._superdomain_scale_factor

    @superdomain_scale_factor.setter
    def superdomain_scale_factor(self, value: FloatLike):
        if (
            not isinstance(value, FloatLike)
            or np.isnan(value)
            or np.isinf(value)
            or value < 1.0
        ):
            raise TypeError(
                "superdomain_scale_factor must be a positive float greater than or equal to 1"
            )
        self._superdomain_scale_factor = value
